"""
Chemical and Disease named entity recognition (NER) evaluation.

This NER test is based on the
Biocreative V Chemical to Disease Relation (BC5CDR)
data, as described below.

Biocreative results here:
https://biocreative.bioinformatics.udel.edu/media/store/files/2015/BC5CDR_overview.final.pdf

In the original task, only disease names were considered in NER evaluation.
(i.e., the task was disease named entity recognition, or DNER).
Across 16 teams, the best systems achieved an F-score of 86.46.
From the paper:
"The average precision, recall and F-score were 78.99%, 74.81% and 76.03%,
respectively. All teams but one achieved a higher F-score than our baseline
dictionary method, which obtained an F-score of 52.30%. While we did not
perform any additional development on DNorm [the benchmark method] to adapt
it to this dataset, it sets a significantly stronger benchmark with a
performance of 80.64% F-score. A total of 7 teams achieved performance
higher than DNorm."

This evaluation also considers chemical named entity recognition.

"""
import gzip
import logging
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from random import shuffle
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

import yaml
from bioc import biocxml
from oaklib import BasicOntologyInterface, get_adapter
from pydantic import BaseModel

from ontogpt.engines.knowledge_engine import chunk_text_by_sentence
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.evaluation.evaluation_engine import SimilarityScore, SPIRESEvaluationEngine
from ontogpt.templates.ctd_ner import (
    ChemicalToDiseaseDocument,
    Chemical,
    Disease,
    Publication,
    TextWithEntity,
)

THIS_DIR = Path(__file__).parent
DATABASE_DIR = Path(__file__).parent / "database"
TEST_SET_BIOC = DATABASE_DIR / "CDR_TestSet.BioC.xml.gz"
TRAIN_SET_BIOC = DATABASE_DIR / "CDR_TrainSet.BioC.xml.gz"

# These are the entity types involved in this dataset.
TARGET_TYPES = ["Chemical", "Disease"]

logger = logging.getLogger(__name__)


class PredictionNER(BaseModel):
    """A prediction for a named entity recognition task."""

    test_object: Optional[TextWithEntity] = None
    """Source of truth to evaluate against."""

    true_positives_ce: Optional[List[Tuple]] = None
    num_true_positives_ce: Optional[int] = None
    false_positives_ce: Optional[List[Tuple]] = None
    num_false_positives_ce: Optional[int] = None
    false_negatives_ce: Optional[List[Tuple]] = None
    num_false_negatives_ce: Optional[int] = None
    true_positives_de: Optional[List[Tuple]] = None
    num_true_positives_de: Optional[int] = None
    false_positives_de: Optional[List[Tuple]] = None
    num_false_positives_de: Optional[int] = None
    false_negatives_de: Optional[List[Tuple]] = None
    num_false_negatives_de: Optional[int] = None
    scores: Optional[Dict[str, SimilarityScore]] = None
    predicted_object: Optional[TextWithEntity] = None
    named_entities: Optional[List[Any]] = None

    def calculate_scores(self, labelers: Optional[List[BasicOntologyInterface]] = None):
        self.scores = {}

        def label(x):
            if labelers:
                for labeler in labelers:
                    lbl = labeler.label(x)
                    if lbl:
                        return f"{x} {lbl}"
            return x

        def all_objects(dm: Optional[TextWithEntity]):
            if dm is not None:
                return list(set(dm.chemicals + dm.diseases))
            else:
                return list()

        def chem_entities(dm: TextWithEntity) -> Set:
            if dm.chemicals is not None:
                return set((label(entity)) for entity in dm.chemicals)
            else:
                return set()

        def disease_entities(dm: TextWithEntity) -> Set:
            if dm.diseases is not None:
                return set((label(entity)) for entity in dm.diseases)
            else:
                return set()

        self.scores["similarity"] = SimilarityScore.from_set(
            all_objects(self.test_object),
            all_objects(self.predicted_object),
            labelers=labelers,
        )

        if self.predicted_object is not None and self.test_object is not None:
            pred_ce = chem_entities(self.predicted_object)
            test_ce = chem_entities(self.test_object)
            pred_de = disease_entities(self.predicted_object)
            test_de = disease_entities(self.test_object)

        self.true_positives_ce = list(pred_ce.intersection(test_ce))
        self.false_positives_ce = list(pred_ce.difference(test_ce))
        self.false_negatives_ce = list(test_ce.difference(pred_ce))
        self.num_false_negatives_ce = len(self.false_negatives_ce)
        self.num_false_positives_ce = len(self.false_positives_ce)
        self.num_true_positives_ce = len(self.true_positives_ce)

        self.true_positives_de = list(pred_de.intersection(test_de))
        self.false_positives_de = list(pred_de.difference(test_de))
        self.false_negatives_de = list(test_de.difference(pred_de))
        self.num_false_negatives_de = len(self.false_negatives_de)
        self.num_false_positives_de = len(self.false_positives_de)
        self.num_true_positives_de = len(self.true_positives_de)


class EvaluationObjectSetNER(BaseModel):
    """A result of performing named entity recognition."""

    precision_ce: float = 0
    recall_ce: float = 0
    f1_ce: float = 0

    precision_de: float = 0
    recall_de: float = 0
    f1_de: float = 0

    training: Optional[List[TextWithEntity]] = None
    predictions: Optional[List[PredictionNER]] = None
    test: Optional[List[TextWithEntity]] = None


@dataclass
class EvalCTDNER(SPIRESEvaluationEngine):
    subject_prefix = "MESH"
    object_prefix = "MESH"

    def __post_init__(self):
        self.extractor = SPIRESEngine(
            template="ctd_ner.ChemicalToDiseaseDocument", model=self.model
        )
        # synonyms are derived entirely from training set
        self.extractor.load_dictionary(DATABASE_DIR / "synonyms.yaml")

    def load_test_cases(self) -> Iterable[ChemicalToDiseaseDocument]:
        return self.load_cases(TEST_SET_BIOC)

    def load_training_cases(self) -> Iterable[ChemicalToDiseaseDocument]:
        return self.load_cases(TRAIN_SET_BIOC)

    def load_cases(self, path: Path) -> Iterable[ChemicalToDiseaseDocument]:
        logger.info(f"Loading {path}")
        with gzip.open(str(path), "rb") as f:
            collection = biocxml.load(f)
            chemicals_by_text = defaultdict(list)
            diseases_by_text = defaultdict(list)
            for document in collection.documents:
                these_annotations = []
                doc = {}
                for p in document.passages:
                    doc[p.infons["type"]] = p.text
                    for a in p.annotations:
                        if a.infons["type"] in TARGET_TYPES:
                            these_annotations.append(a)
                title = doc["title"]
                abstract = doc["abstract"]
                logger.debug(f"Title: {title} Abstract: {abstract}")
                for a in these_annotations:
                    i = a.infons
                    if i["type"] == "Chemical":
                        e = Chemical.model_validate(
                            {
                                "id": f"{self.subject_prefix}:{i[self.subject_prefix]}",
                            }
                        )
                        chemicals_by_text[(title, abstract)].append(e.id)
                    elif i["type"] == "Disease":
                        e = Disease.model_validate(
                            {
                                "id": f"{self.subject_prefix}:{i[self.subject_prefix]}",
                            }
                        )
                        diseases_by_text[(title, abstract)].append(e.id)

        all_entities_by_text = chemicals_by_text | diseases_by_text

        i = 0
        for (title, abstract), _entities in all_entities_by_text.items():
            i = i + 1
            pub = Publication.model_validate(
                {
                    "id": str(i),
                    "title": title,
                    "abstract": abstract,
                }
            )
            chemical_entities = chemicals_by_text[(title, abstract)]
            disease_entities = diseases_by_text[(title, abstract)]
            logger.debug(
                f"Chemicals: {len(chemical_entities)} for Title: {title} Abstract: {abstract}"
            )
            logger.debug(
                f"Diseases: {len(disease_entities)} for Title: {title} Abstract: {abstract}"
            )
            yield ChemicalToDiseaseDocument.model_validate(
                {
                    "publication": pub,
                    "chemicals": chemical_entities,
                    "diseases": disease_entities,
                }
            )

    def create_training_set(self, num=100):
        ke = self.extractor
        docs = list(self.load_test_cases())
        shuffle(docs)
        for doc in docs[0:num]:
            text = doc.text
            prompt = ke.get_completion_prompt(None, text)
            completion = ke.serialize_object()
            yield dict(prompt=prompt, completion=completion)

    def eval(self) -> EvaluationObjectSetNER:
        """Evaluate the ability to extract chemical and disease named entities."""
        labeler = get_adapter("sqlite:obo:mesh")
        if self.num_tests and isinstance(self.num_tests, int):
            num_test = self.num_tests
        else:
            num_test = 1
        ke = self.extractor
        docs = list(self.load_test_cases())
        shuffle(docs)
        eos = EvaluationObjectSetNER(
            test=docs[:num_test],
            training=[],
            predictions=[],
        )
        n = 1
        for doc in eos.test:
            logger.info(f"Iteration {n} of {num_test}")
            n += 1
            logger.info(doc)
            text = f"Title: {doc.publication.title} Abstract: {doc.publication.abstract}"
            pub = Publication.model_validate(
                {
                    "id": str(doc.publication.id),
                    "title": doc.publication.title,
                    "abstract": doc.publication.abstract,
                }
            )
            predicted_obj = ChemicalToDiseaseDocument.model_validate(
                {
                    "publication": pub,
                    "chemicals": [],
                    "diseases": [],
                }
            )
            named_entities: List[str] = []  # This stores the NEs for the whole document
            ke.named_entities = []  # This stores the NEs the extractor knows about

            if self.chunking:
                text_list = chunk_text_by_sentence(text)
            else:
                text_list = iter([text])

            for chunked_text in text_list:
                extraction = ke.extract_from_text(chunked_text)
                if extraction.extracted_object is not None:
                    logger.info(
                        f"{len(extraction.extracted_object.chemicals)}\
                            chemical entities from window: {chunked_text}"
                    )
                    logger.info(
                        f"{len(extraction.extracted_object.diseases)}\
                            disease entities from window: {chunked_text}"
                    )
                if not predicted_obj and extraction.extracted_object is not None:
                    predicted_obj = extraction.extracted_object
                else:
                    if predicted_obj is not None and extraction.extracted_object is not None:
                        predicted_obj.chemicals.extend(extraction.extracted_object.chemicals)
                        predicted_obj.diseases.extend(extraction.extracted_object.diseases)
                        logger.info(
                            f"{len(predicted_obj.chemicals)} chemical entities, after concatenation"
                        )
                        logger.info(
                            f"{len(predicted_obj.diseases)} disease entities, after concatenation"
                        )
                        logger.debug(f"concatenated chemical entities: {predicted_obj.chemicals}")
                        logger.debug(f"concatenated disease entities: {predicted_obj.diseases}")
                if extraction.named_entities is not None:
                    for entity in extraction.named_entities:
                        if entity not in named_entities:
                            named_entities.append(entity)

            def included(t: str):
                if t.startswith("MESH:"):
                    return t

            predicted_obj.chemicals = [t for t in predicted_obj.chemicals if included(t)]
            predicted_obj.diseases = [t for t in predicted_obj.diseases if included(t)]

            logger.info(f"{len(predicted_obj.chemicals)} filtered chemical entities (MESH only)")
            logger.info(f"{len(predicted_obj.diseases)} filtered disease entities (MESH only)")
            pred = PredictionNER(
                predicted_object=predicted_obj, test_object=doc, named_entities=named_entities
            )
            named_entities.clear()
            logger.info("PRED")
            logger.info(yaml.dump(data=pred.model_dump()))
            logger.info("Calc scores")
            pred.calculate_scores(labelers=[labeler])
            logger.info(yaml.dump(data=pred.model_dump()))
            eos.predictions.append(pred)
        self.calc_stats(eos)
        return eos

    def calc_stats(self, eos: EvaluationObjectSetNER):
        num_true_positives_ce = sum(p.num_true_positives_ce for p in eos.predictions)
        num_false_positives_ce = sum(p.num_false_positives_ce for p in eos.predictions)
        num_false_negatives_ce = sum(p.num_false_negatives_ce for p in eos.predictions)
        if num_true_positives_ce + num_false_positives_ce == 0:
            logger.warning("No true positives or false positives for chemical entities.")
            return
        eos.precision_ce = num_true_positives_ce / (num_true_positives_ce + num_false_positives_ce)
        eos.recall_ce = num_true_positives_ce / (num_true_positives_ce + num_false_negatives_ce)
        if eos.precision_ce + eos.recall_ce == 0:
            logger.warning("No precision or recall for chemical entities.")
            return
        eos.f1_ce = 2 * (eos.precision_ce * eos.recall_ce) / (eos.precision_ce + eos.recall_ce)

        num_true_positives_de = sum(p.num_true_positives_de for p in eos.predictions)
        num_false_positives_de = sum(p.num_false_positives_de for p in eos.predictions)
        num_false_negatives_de = sum(p.num_false_negatives_de for p in eos.predictions)
        if num_true_positives_de + num_false_positives_de == 0:
            logger.warning("No true positives or false positives for disease entities.")
            return
        eos.precision_de = num_true_positives_de / (num_true_positives_de + num_false_positives_de)
        eos.recall_de = num_true_positives_de / (num_true_positives_de + num_false_negatives_de)
        if eos.precision_de + eos.recall_de == 0:
            logger.warning("No precision or recall for disease entities.")
            return
        eos.f1_de = 2 * (eos.precision_de * eos.recall_de) / (eos.precision_de + eos.recall_de)



"""CSV exporter class."""

from dataclasses import dataclass
from io import BytesIO
from pathlib import Path
from typing import Optional, TextIO, Union
from ontogpt.io.exporter import Exporter
from io import BytesIO, StringIO, TextIOWrapper
import pandas as pd

from linkml_runtime import SchemaView

from ontogpt.templates.core import ExtractionResult


@dataclass
class CSVExporter(Exporter):

    sep: str = ","

    def export(
        self,
        extraction_output: ExtractionResult,
        output: Union[str, Path, TextIO, BytesIO],
        schemaview: Optional[SchemaView],
    ):
        if isinstance(output, Path):
            output = open(str(output), "w", encoding="utf-8")
        if isinstance(output, str):
            output = StringIO(output)
        if isinstance(output, BytesIO):
            output = TextIOWrapper(output, encoding="utf-8")
        out_dict = extraction_output.extracted_object.dict()
        df = pd.DataFrame.from_dict(out_dict, orient="index")
        df.columns = ["values"]
        df = df.explode("values")
        df.to_csv(path_or_buf=output, sep=self.sep, header=True)



"""ontogpt package."""

from importlib import metadata
from pathlib import Path

from litellm import get_model_cost_map

rel_path = Path(__file__).resolve()

# Define the default model
DEFAULT_MODEL = "gpt-4o"

# Define the default temperature
# This assumes the OpenAI default, which is between 0 and 2
DEFAULT_TEMPERATURE = 1.0

# Define the default embedding model
DEFAULT_EMBEDDING_MODEL = "text-embedding-ada-002"

# These input formats are used in the CLI
VALID_INPUT_FORMATS = [".csv", ".tsv", ".txt", ".od", ".odf", ".ods", ".pdf", ".xls", ".xlsx"]
VALID_TABULAR_FORMATS = [".csv", ".tsv"]
VALID_SPREADSHEET_FORMATS = [".od", ".odf", ".ods", ".xls", ".xlsb", ".xlsm", ".xlsx"]

# Build list of available models
# This is provided by the litellm package
MODELS = get_model_cost_map("")

try:
    __version__ = metadata.version(__name__)
except metadata.PackageNotFoundError:
    # package is not installed
    __version__ = "0.0.0"  # pragma: no cover



"""OntoGPT Webapp main function.

This module sets up a FastAPI web application for
the OntoGPT project. It includes routes for handling
GET and POST requests, and integrates with components
such as SPIRESEngine and HTMLExporter to process and
display data.

Classes:
    Query: A Pydantic model representing the structure
    of the form data.

Functions:
    get_engine(datamodel: str, llm_model: str) -> SPIRESEngine:
        Retrieves or initializes a SPIRESEngine instance for
        the given datamodel and LLM model.

    read_root(request: Request) -> TemplateResponse:
        Handles GET requests to the root URL,
        rendering the form.html template.

    form_post(request: Request, datamodel: str = Form(...),
        text: str = Form(...), llm_model: str = Form(...))
        -> TemplateResponse:
        Handles POST requests to the root URL, processing
        the form data and rendering the results.html template.

    start():
        Launches the FastAPI application using Uvicorn.
"""

from io import StringIO
from pathlib import Path
from typing import Dict

import uvicorn
from fastapi import FastAPI, Form, Request
from pydantic import BaseModel
from starlette.staticfiles import StaticFiles
from starlette.templating import Jinja2Templates

from ontogpt.cli import _get_templates
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.io.html_exporter import HTMLExporter
from ontogpt.io.template_loader import get_template_details

this_path = Path(__file__).parent
static_dir = this_path / "static"
html_dir = this_path / "html"

# Populate the dict of available models
DATAMODELS = {}
all_templates = _get_templates()
all_templates = dict(sorted(all_templates.items()))
for template_id, (_name, description) in all_templates.items():
    DATAMODELS[template_id] = f"{description}"

LLM_MODELS = [
    "gpt-4o-mini",
    "gpt-4o",
    "gpt-4",
    "gpt-4-turbo",
    "ollama/llama2",
    "ollama/llama3",
    "ollama/orca-mini",
]


class Query(BaseModel):
    text: str
    datamodel: str
    model: str


app = FastAPI()
static_dir.mkdir(parents=True, exist_ok=True)
app.mount("/static", StaticFiles(directory=str(static_dir)), name="static")
templates = Jinja2Templates(directory=str(html_dir))

html_exporter = HTMLExporter(output=None)
engines: Dict[str, SPIRESEngine] = {}


def get_engine(datamodel: str, llm_model: str):
    if datamodel not in engines:
        template_details = get_template_details(template=datamodel)
        engines[datamodel] = SPIRESEngine(model=llm_model, template_details=template_details)
    return engines[datamodel]


@app.get("/")
def read_root(request: Request):
    return templates.TemplateResponse(
        "form.html",
        context={"request": request, "datamodels": DATAMODELS, "llm_models": LLM_MODELS},
    )


@app.post("/")
def form_post(
    request: Request, datamodel: str = Form(...), text: str = Form(...), llm_model: str = Form(...)
):
    print(f"Received request with schema {datamodel}")
    print(f"Received request with text {text}")
    print(f"Received request to be sent to {llm_model}")
    engine = get_engine(datamodel, llm_model)
    ann = engine.extract_from_text(text)
    print(f"Got {ann}")
    output = StringIO()
    html_exporter.export(extraction_output=ann, output=output)
    return templates.TemplateResponse(
        "results.html", context={"request": request, "inner_html": output.getvalue()}
    )


def start():
    """Launch with `uv run start` at root level."""
    uvicorn.run("ontogpt.webapp.main:app", host="127.0.0.1", port=8000, reload=True)



"""
Topic classifier engine.

This module contains the `TopicClassifierEngine` class, which inherits from `KnowledgeEngine`.
The engine is designed to classify input text based on a given topic.

Classes:
    TopicClassifierEngine: Engine for classifying input text based on its topic.

Methods:
    binary_classify(topic: str, text: str) -> bool:
        Given a topic description, indicate whether it applies to the input text.
        Returns True if the text matches the topic, otherwise returns False.
"""

import logging

from dataclasses import dataclass

from ontogpt.engines.knowledge_engine import KnowledgeEngine

logger = logging.getLogger(__name__)


@dataclass
class TopicClassifierEngine(KnowledgeEngine):
    """Engine for classifying input text based on its topic."""

    def binary_classify(self, topic: str, text: str) -> bool:
        """Given a topic description, indicate whether it applies to the input text."""

        prompt = (
            "I will provide a topic followed by a text."
            " If the text matches the topic, return only 'True'."
            " If the text does not match the topic, return only 'False'."
            " Do not provide any other text."
            f" Topic: {topic}"
            f" Text: {text}"
        )
        payload = self.client.complete(prompt)
        logger.info(f"Binary classification result: {payload}")
        if payload.strip().lower() in ["true"]:
            return True
        else:
            return False



"""JSON Wrapper."""

import io
import logging
import json
from typing import Any, Optional, TextIO

from ontogpt.io.utils import eliminate_empty


logger = logging.getLogger(__name__)


def dump_minimal_json(obj: Any, minimize=True, file: Optional[TextIO] = None) -> str:
    """Dump a JSON string, but eliminate Nones and empty lists and dicts."""
    if not file:
        file = io.StringIO()
        json.dump(eliminate_empty(obj, not minimize), file, indent=2)
        return file.getvalue()
    else:
        json.dump(eliminate_empty(obj, not minimize), file, indent=2)
        return ""



"""Utility for running multilingual analysis."""

import logging
import os
from typing import Tuple

from ontogpt.clients import LLMClient
from ontogpt.engines.spires_engine import SPIRESEngine
from ontogpt.io.template_loader import get_template_details
from ontogpt.templates.core import ExtractionResult


def multilingual_analysis(
    prompt,
    filename,
    output_directory,
    model,
    temperature,
    api_base,
    api_version,
    model_provider,
    system_message,
) -> Tuple[ExtractionResult, str, SPIRESEngine]:
    """Run the multilingual analysis."""
    # Set up the extraction template
    template = "all_disease_grounding"
    template_details = get_template_details(template=template)

    ai = LLMClient()
    ai.model = model

    # Keep track of the predictions
    # Key is the filename, value is a list of predictions
    pred_ids: dict[str, list[str]] = {}
    pred_names: dict[str, list[str]] = {}

    try:
        gpt_diagnosis = ai.complete(prompt)
        print(gpt_diagnosis)
    except Exception as e:
        raise Exception(f"Error in completion: {e}")

    # Call the extract function here
    # to ground the answer to OMIM (using MONDO, etc)
    # The KE is refreshed here to avoid retaining
    ke = SPIRESEngine(
        template_details=template_details,
        model=model,
        temperature=temperature,
        api_base=api_base,
        api_version=api_version,
        model_provider=model_provider,
        system_message=system_message,
    )
    extraction = ke.extract_from_text(text=gpt_diagnosis)
    predictions = extraction.named_entities
    pred_ids[filename] = []
    pred_names[filename] = []
    for pred in predictions:
        pred_ids[filename].append(pred.id)
        pred_names[filename].append(pred.label)

    if len(pred_ids[filename]) == 0:
        raise Exception(f"No grounded IDs found for {filename}")

    # Log the result
    logging.info("input file name" "\tpredicted diagnosis ids\tpredicted diagnosis names\n")
    logging.info(
        f"{filename}" f'\t{"|".join(pred_ids[filename])}' f'\t{"|".join(pred_names[filename])}\n'
    )

    # Retain the output as an ExtractionResult object
    # Create the output filename based on the input filename
    output_file_name = filename + ".result"
    output_file_path = os.path.join(output_directory, output_file_name)
    extraction.extracted_object.label = filename

    return (extraction, output_file_path, ke)



"""
Synonym engine.

This module defines the SynonymEngine class,
which is responsible for generating synonyms
for given named entities within a specified domain.

Classes:
    SynonymEngine: Inherits from KnowledgeEngine and
    provides a method to retrieve synonyms.

Methods:
    synonyms(named_entity: str, domain: str) -> List[str]:
        Retrieves synonyms for the given named entity
        within the specified domain.
"""

from dataclasses import dataclass
from typing import List

from ontogpt.engines.knowledge_engine import KnowledgeEngine


@dataclass
class SynonymEngine(KnowledgeEngine):
    """Engine for generating synonyms."""

    def synonyms(self, named_entity: str, domain: str) -> List[str]:
        """Get synonyms for a given text."""
        prompt = f"List the example formal scientific\
            synonyms for the {domain} concept {named_entity} as a semi-colon separated list."
        prompt += " Only include terms with identical meaning, not more specific or general terms."
        payload = self.client.complete(prompt)
        return payload.split("; ")



from __future__ import annotations 
from datetime import (
    datetime,
    date
)
from decimal import Decimal 
from enum import Enum 
import re
import sys
from typing import (
    Any,
    List,
    Literal,
    Dict,
    Optional,
    Union
)
from pydantic.version import VERSION  as PYDANTIC_VERSION 
if int(PYDANTIC_VERSION[0])>=2:
    from pydantic import (
        BaseModel,
        ConfigDict,
        Field,
        field_validator
    )
else:
    from pydantic import (
        BaseModel,
        Field,
        validator
    )

metamodel_version = "None"
version = "None"


class ConfiguredBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment = True,
        validate_default = True,
        extra = "forbid",
        arbitrary_types_allowed = True,
        use_enum_values = True,
        strict = False,
    )
    pass


class NullDataOptions(str, Enum):
    UNSPECIFIED_METHOD_OF_ADMINISTRATION = "UNSPECIFIED_METHOD_OF_ADMINISTRATION"
    NOT_APPLICABLE = "NOT_APPLICABLE"
    NOT_MENTIONED = "NOT_MENTIONED"


class ExtractionResult(ConfiguredBaseModel):
    """
    A result of extracting knowledge on text
    """
    input_id: Optional[str] = Field(None)
    input_title: Optional[str] = Field(None)
    input_text: Optional[str] = Field(None)
    raw_completion_output: Optional[str] = Field(None)
    prompt: Optional[str] = Field(None)
    extracted_object: Optional[Any] = Field(None, description="""The complex objects extracted from the text""")
    named_entities: Optional[List[Any]] = Field(default_factory=list, description="""Named entities extracted from the text""")


class NamedEntity(ConfiguredBaseModel):
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class CompoundExpression(ConfiguredBaseModel):
    pass


class Triple(CompoundExpression):
    """
    Abstract parent for Relation Extraction tasks
    """
    subject: Optional[str] = Field(None)
    predicate: Optional[str] = Field(None)
    object: Optional[str] = Field(None)
    qualifier: Optional[str] = Field(None, description="""A qualifier for the statements, e.g. \"NOT\" for negation""")
    subject_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the subject of the statement, e.g. \"high dose\" or \"intravenously administered\"""")
    object_qualifier: Optional[str] = Field(None, description="""An optional qualifier or modifier for the object of the statement, e.g. \"severe\" or \"with additional complications\"""")


class TextWithTriples(ConfiguredBaseModel):
    """
    A text containing one or more relations of the Triple type.
    """
    publication: Optional[Publication] = Field(None)
    triples: Optional[List[Triple]] = Field(default_factory=list)


class TextWithEntity(ConfiguredBaseModel):
    """
    A text containing one or more instances of a single type of entity.
    """
    publication: Optional[Publication] = Field(None)
    entities: Optional[List[str]] = Field(default_factory=list)


class RelationshipType(NamedEntity):
    id: str = Field(..., description="""A unique identifier for the named entity""")
    label: Optional[str] = Field(None, description="""The label (name) of the named thing""")


class Publication(ConfiguredBaseModel):
    id: Optional[str] = Field(None, description="""The publication identifier""")
    title: Optional[str] = Field(None, description="""The title of the publication""")
    abstract: Optional[str] = Field(None, description="""The abstract of the publication""")
    combined_text: Optional[str] = Field(None)
    full_text: Optional[str] = Field(None, description="""The full text of the publication""")


class AnnotatorResult(ConfiguredBaseModel):
    subject_text: Optional[str] = Field(None)
    object_id: Optional[str] = Field(None)
    object_text: Optional[str] = Field(None)


class STORMSChecklist(ConfiguredBaseModel):
    """
    The complete STORMS checklist.
    """
    abstract_structured_or_unstructured: Optional[str] = Field(None, description="""Abstract should include information on background, methods, results, and conclusions in structured or unstructured format.""")
    abstract_study_design: Optional[str] = Field(None, description="""State study design in abstract.""")
    abstract_sequencing_methods: Optional[str] = Field(None, description="""State the strategy used for metagenomic classification.""")
    abstract_specimens: Optional[str] = Field(None, description="""Describe body site(s) studied.""")
    introduction_background_and_rationale: Optional[str] = Field(None, description="""Summarize the underlying background, scientific evidence, or theory driving the current hypothesis as well as the study objectives.""")
    introduction_hypotheses: Optional[str] = Field(None, description="""State the pre-specified hypothesis. If the study is exploratory, state any pre-specified study objectives.""")
    methods_study_design: Optional[str] = Field(None, description="""Describe the study design.""")
    methods_participants: Optional[str] = Field(None, description="""State what the population of interest is, and the method by which participants are sampled from that population.""")
    methods_geographic_location: Optional[str] = Field(None, description="""State the geographic region(s) where participants were sampled from.""")
    methods_relevant_dates: Optional[str] = Field(None, description="""State the start and end dates for recruitment, follow-up, and data collection.""")
    methods_eligibility_criteria: Optional[str] = Field(None, description="""List any criteria for inclusion and exclusion of recruited participants.""")
    methods_antibiotics_usage: Optional[str] = Field(None, description="""List what is known about antibiotics usage before or during sample collection.""")
    methods_analytic_sample_size: Optional[str] = Field(None, description="""Explain how the final analytic sample size was calculated, including the number of cases and controls if relevant, and reasons for dropout at each stage of the study.""")
    methods_longitudinal_studies: Optional[str] = Field(None, description="""For longitudinal studies, state how many follow-ups were conducted, describe sample size at follow-up by group or condition, and discuss any loss to follow-up.""")
    methods_matching: Optional[str] = Field(None, description="""For matched studies, give matching criteria.""")
    methods_ethics: Optional[str] = Field(None, description="""State the name of the institutional review board that approved the study and protocols, protocol number and date of approval, and procedures for obtaining informed consent from participants.""")
    methods_laboratory_methods: Optional[str] = Field(None, description="""State the laboratory/center where laboratory work was done.""")
    methods_specimen_collection: Optional[str] = Field(None, description="""State the body site(s) sampled from and how specimens were collected.""")
    methods_shipping: Optional[str] = Field(None, description="""Describe how samples were stored and shipped to the laboratory.""")
    methods_storage: Optional[str] = Field(None, description="""Describe how the laboratory stored samples, including time between collection and storage and any preservation buffers or refrigeration used.""")
    methods_dna_extraction: Optional[str] = Field(None, description="""Provide DNA extraction method, including kit and version if relevant.""")
    methods_human_dna_sequence_depletion_or_microbial_dna_enrichment: Optional[str] = Field(None, description="""Describe whether human DNA sequence depletion or enrichment of microbial or viral DNA was performed.""")
    methods_primer_selection: Optional[str] = Field(None, description="""Provide primer selection and DNA amplification methods as well as variable region sequenced (if applicable).""")
    methods_positive_controls: Optional[str] = Field(None, description="""Describe any positive controls (mock communities) if used.""")
    methods_negative_controls: Optional[str] = Field(None, description="""Describe any negative controls if used.""")
    methods_contaminant_mitigation_and_identification: Optional[str] = Field(None, description="""Provide any laboratory or computational methods used to control for or identify microbiome contamination from the environment, reagents, or laboratory.""")
    methods_replication: Optional[str] = Field(None, description="""Describe any biological or technical replicates included in the sequencing, including which steps were replicated between them.""")
    methods_sequencing_strategy: Optional[str] = Field(None, description="""Major divisions of strategy, such as shotgun or amplicon sequencing.""")
    methods_sequencing_methods: Optional[str] = Field(None, description="""State whether experimental quantification was used (QMP/cell count based, spike-in based) or whether relative abundance methods were applied.""")
    methods_batch_effects: Optional[str] = Field(None, description="""Detail any blocking or randomization used in study design to avoid confounding of batches with exposures or outcomes. Discuss any likely sources of batch effects, if known.""")
    methods_metatranscriptomics: Optional[str] = Field(None, description="""Detail whether any mRNA enrichment was performed and whether/how retrotranscription was performed prior to sequencing. Provide size range of isolated transcripts. Describe whether the sequencing library was stranded or not. Provide details on sequencing methods and platforms.""")
    methods_metaproteomics: Optional[str] = Field(None, description="""Detail which protease was used for digestion. Provide details on proteomic methods and platforms (e.g. LC-MS/MS, instrument type, column type, mass range, resolution, scan speed, maximum injection time, isolation window, normalised collision energy, and resolution).""")
    methods_metabolomics: Optional[str] = Field(None, description="""Specify the analytic method used (such as nuclear magnetic resonance spectroscopy or mass spectrometry). For mass spectrometry, detail which fractions were obtained (polar and/or non polar) and how these were analyzed. Provide details on metabolomics methods and platforms (e.g. derivatization, instrument type, injection type, column type and instrument settings).""")
    results_descriptive_data: Optional[str] = Field(None, description="""Give characteristics of study participants (e.g. dietary, demographic, clinical, social) and information on exposures and potential confounders.""")
    results_microbiome_data: Optional[str] = Field(None, description="""Report descriptive findings for microbiome analyses with all applicable outcomes and covariates.""")
    results_taxonomy: Optional[str] = Field(None, description="""Identify taxonomy using standardized taxon classifications that are sufficient to uniquely identify taxa.""")
    results_differential_abundance: Optional[str] = Field(None, description="""Report results of differential abundance analysis by the variable of interest and (if applicable) by time, clearly indicating the direction of change and total number of taxa tested.""")
    results_other_data_types: Optional[str] = Field(None, description="""Report other data analyzed--e.g. metabolic function, functional potential, MAG assembly, and RNAseq.""")
    results_other_statistical_analysis: Optional[str] = Field(None, description="""Report any statistical data analysis not covered above.""")
    discussion_key_results: Optional[str] = Field(None, description="""Summarize key results with reference to study objectives.""")
    discussion_interpretation: Optional[str] = Field(None, description="""Give a cautious overall interpretation of results considering objectives, limitations, multiplicity of analyses, results from similar studies, and other relevant evidence.""")
    discussion_limitations: Optional[str] = Field(None, description="""Discuss limitations of the study, taking into account sources of potential bias or imprecision.""")
    discussion_bias: Optional[str] = Field(None, description="""Discuss any potential for bias to influence study findings.""")
    discussion_generalizability: Optional[str] = Field(None, description="""Discuss the generalizability (external validity) of the study results.""")
    discussion_ongoing_future_work: Optional[str] = Field(None, description="""Describe potential future research or ongoing research based on the study's findings.""")
    other_information_funding: Optional[str] = Field(None, description="""Give the source of funding and the role of the funders for the present study and, if applicable, for the original study on which the present article is based.""")
    other_information_acknowledgements: Optional[str] = Field(None, description="""Include acknowledgements of those who contributed to the research but did not meet criteria for authorship.""")
    other_information_conflicts_of_interest: Optional[str] = Field(None, description="""Include a conflicts of interest statement.""")
    other_information_supplements: Optional[str] = Field(None, description="""Indicate where supplements may be accessed and what materials they contain.""")
    other_information_supplementary_data: Optional[str] = Field(None, description="""Provide supplementary data files of results with all taxa and all outcome variables analyzed. Indicate the taxonomic level of all taxa.""")


# Model rebuild
# see https://pydantic-docs.helpmanual.io/usage/models/#rebuilding-a-model
ExtractionResult.model_rebuild()
NamedEntity.model_rebuild()
CompoundExpression.model_rebuild()
Triple.model_rebuild()
TextWithTriples.model_rebuild()
TextWithEntity.model_rebuild()
RelationshipType.model_rebuild()
Publication.model_rebuild()
AnnotatorResult.model_rebuild()
STORMSChecklist.model_rebuild()




"""IO Utilities."""

from typing import Any

import pydantic


def eliminate_empty(obj: Any, preserve=False) -> Any:
    """Eliminate empty lists and dicts from an object."""
    if isinstance(obj, list):
        return [eliminate_empty(x, preserve) for x in obj if x or preserve]
    elif isinstance(obj, dict):
        return {k: eliminate_empty(v, preserve) for k, v in obj.items() if v or preserve}
    elif isinstance(obj, pydantic.BaseModel):
        return eliminate_empty(obj.model_dump(), preserve)
    elif isinstance(obj, tuple):
        return [eliminate_empty(x, preserve) for x in obj]
    elif isinstance(obj, str):
        return str(obj)
    else:
        return obj
